% dfan_lqr.m - control LQR de un ventilador

%%
%% Dinamica de un ventilador de flujo guiado (tipo de propulsor)
%%
%% Esta es la dinamica de un ventilador propulsor, escrita en la forma de espacio de 
%% estados. Basicamente es un ejemplo parecido a la aeronave de impulsion 
%% Harriet, excepto por la inclusion del termino de excentricidad l 
%% diferente de cero.
%%

close all;
clear all;
clc;

% Parametros del sistema
J = 0.0475;			% inertia around pitch axis
m = 1.5;			% mass of fan
r = 0.25;			% distance to flaps
g = 10;				% gravitational constant
gamma = 0.51;		% counterweighted mass
d = 0.2;	 		% damping factor (estimated)
l = 0.05;			% offset of center of mass

% System matrices (entire plant: 2 input, 2 output)
A = [  0     0     0     1     0     0;
       0     0     0     0     1     0;
       0     0     0     0     0     1;
       0     0  -gamma -d/m    0     0;
       0     0     0     0   -d/m    0;
       0     0 -m*g*l/J  0     0     0 ];
 
B = [  0     0;
       0     0;
       0     0;
      1/m    0;
       0    1/m;
      r/J    0    ];

C = [  1     0    0      0     0     0;
       0     1    0      0     0     0 ];

D = [  0     0;   0      0];

%%
%% Construccion de entradas y salidas correspondientes a escalones en
%% posiciones xy
%%
%% Los vectores xd y yd corresponden a los es We
%% construct the control law by computing an LQR gain for the augmented
%% system.tados de equilibrio del 
%% sistema deseados. Las matrices Cx y Cy son las correspondientes salidas.
%%
%% La forma en que estos vectores son usados es para calcular la dinamica
%% del sistema en lazo cerrado
%%
%%	xdot = Ax + B u		=>	xdot = (A-BK)x + K xd
%%         u = -K(x - xd)		   y = Cx
%%
%% La dinamica en lazo cerrado puede ser simulada usando la funcion "step", 
%% con K*xd como un vector de entradas (asuma que la entrada es de tamano unitario,
%% tal que xd corresponde al estado en estado estacionario deseado.
%%

xd = [1; 0; 0; 0; 0; 0];  Cx = [1 0 0 0 0 0];
yd = [0; 1; 0; 0; 0; 0];  Cy = [0 1 0 0 0 0];

%%
%% Diseno LQR
%%

% Comenzar con una matriz diagonal
Qx1 = diag([1, 1, 1, 1, 1, 1]);
Qu1a = 0.1 * diag([1, 1]);
K1a = lqr(A, B, Qx1, Qu1a);

% Cerrar el lazo de control: xdot = Ax + B K (x-xd)
H1ax = ss(A-B*K1a,B(:,1)*K1a(1,:)*xd,Cx,0);
H1ay = ss(A-B*K1a,B(:,2)*K1a(2,:)*yd,Cy,0);

figure(1)
step(H1ax, H1ay, 10); title('');
lgh = legend('x', 'y'); legend(lgh, 'boxoff');
%print -depsc 'dfan-lqrstep1.eps'; % para obtener archivo
% tipo eps

% Observar diferencias matrices de ponderacion R
Qu1b = diag([1, 1]); K1b = lqr(A, B, Qx1, Qu1b);
H1bx = ss(A-B*K1b,B(:,1)*K1b(1,:)*xd,Cx,0);

Qu1c = diag([10, 10]); K1c = lqr(A, B, Qx1, Qu1c);
H1cx = ss(A-B*K1c,B(:,1)*K1c(1,:)*xd,Cx,0);

figure(2)
step(H1ax, H1bx, H1cx, 10); title('');
lgh = legend('rho = 0.1', 'rho = 1', 'rho = 10', 'Location', 'southeast');
legend(lgh, 'boxoff');
%print -depsc 'dfun-lqrstep2.eps'; end;

% Ponderando la salida
Qx2 = [Cx; Cy]' * [Cx; Cy];
Qu2 = 0.1 * diag([1, 1]);
K2 = lqr(A, B, Qx2, Qu2);

H2x = ss(A-B*K2,B(:,1)*K2(1,:)*xd,Cx,0);
H2y = ss(A-B*K2,B(:,2)*K2(2,:)*yd,Cy,0);

figure(3); step(H2x, H2y, 10);
lgh = legend('x', 'y'); 

%%
%% Ponderacion con motivacion fisica
%%
%% Si deseamos tener un error de 1 cm en x, un error de 10 cm en y.
%% Intentando mantener el angulo a menos de 5 grados durante los ajustes.
%% Penalizar las fuerzas debido a una perdida de eficiencia
%%

Qx3 = diag([100, 1, 2*pi/9, 0, 0, 0]);
Qu3 = diag([10, 1]);
K3 = lqr(A, B, Qx3, Qu3);

H3x = ss(A-B*K3,B(:,1)*K3(1,:)*xd,Cx,0);
H3y = ss(A-B*K3,B(:,2)*K3(2,:)*yd,Cy,0);
[x, Tx, states] = step(H3x, 10);
u1 = K3(1,:) * states';

[y, Ty, states] = step(H3y, 10);
u2 = K3(2,:) * states';

% First plot the x and y position
figure(4)
plot(Tx, x, '-', Ty, y, '--');

lgh = legend('x', 'y'); legend(lgh, 'boxoff');
%print -depsc 'dfun-lqr-phystates.eps'; 

% Now print the inputs
figure(5)
plot(Tx, u1, '-', Ty, u2, '--');

lgh = legend('u1', 'u2'); legend(lgh, 'boxoff');
% print -depsc 'dfan-lqr-phyinputs.eps'; 

%%
%% Control de Velocidad
%%
%% En este ejemplo, modificamos el sistema tal que podamos controlar la 
%% velocidad en la direccion x. Ignorando la dinamica en la direccion 
%% vertical (y). Esta dinamica demuestra el rol del sistema de 
%% realimentacion directa dado que el punto de equilibrio correspondiente a
%% vd diferente de cero requiere una entrada diferente de cero.
%%
%% Para este ejemplo, usamos la ley de control u = -K(x-xd) + ud y
%% convertimos a la forma u = -K x + N r, donde r es la entrada de
%% referencia y N es calculado como a continuacion (equivalente a descrito en clase).
%%

% Extrayendo la dinamica: theta, xdot, thdot
Av = A([3 4 6], [3 4 6]);
Bv = B([3 4 6], 1);
Cv = [0 1 0];				% eligiendo vx como salida
Dv = 0;	

% Derivar la matriz de ganancias K usando LQR
Qxv = diag([2*pi/5, 10, 0]);
Quv = 0.1;
Kv = lqr(Av, Bv, Qxv, Quv);

% Disenar el termino de realimentacion directa kr 
T = [Av Bv; Cv Dv];			% system matrix
Nxu = T \ [0; 0; 0; 1];			% compute [Nx; Nu]
Nx = Nxu(1:3); Nu = Nxu(4);		% extract Nx and Nu
N = Nu + Kv*Nx;				% compute feedforward term

%%
%% Design #1: Sin alimentacion directa, ud
%%

Nv1 = [0; 1; 0];
Hv1 = ss(Av-Bv*Kv, Bv*Kv*Nx, Cv, 0);
figure(6); step(Hv1, 10);

%%
%% Design #2: Calculando ganancia de alimentacion directa correspondiente
%% al punto de equilibrio 
%%
Hv2 = ss(Av-Bv*Kv, Bv*N, Cv, 0);
figure(7); step(Hv2, 10);

%%
%% Design #3: Accion integral
%%
%% Aumentando un estado al sistema tal que xidot = v - vd. Se construye la
%% ley de control mediante calculo de la ganancia LQR para el sistema aumentado
%%

Ai = [Av, [0; 0; 0]; [Cv, 0]];
Bi = [Bv; 0];
Ci = [Cv, 0];
Di = Dv;

% Disenar el termino de realimentacion, incluyendo ponderacion en la
% integral del error
Qxi = diag([2*pi/5, 10, 0, 10]);	% Try setting Qxi4 to 1, 0.1, etc
Qui = 0.1;
Ki = lqr(Ai, Bi, Qxi, Qui);

% Estados con integral del error (estados aumentado)
xid = [0; 1; 0; 0];

% Construir el sistema en lazo cerrado (incluyendo integrador)
Hi = ss(Ai-Bi*Ki,Bi*Ki*xid - [0; 0; 0; Ci*xid],Ci,0);
figure(8); step(Hi, 10);
