% pvtol.m - modelo no lineal ventilador de flujo direccionado, con LQR y EKF
% RMM, 5 Feb 06
%
% Esta funcion contiene la dinamica para un ventilador con flujo
% direccionado, con un controlador LQR y estimador FK.
%
% state(1)		x position, in meters
% state(2)		y position, in meters
% state(3)		theta angle, in radians
% state(4-6)	velocities
% state(7-12)	estimated states
% state(13-48)	covariance matrix (ordered rowise)

function deriv = pvtol(t, state, flags)
global pvtol_K;				% LQR gain
global pvtol_L;				% LQE gain (temporary)
global pvtol_Rv;			% Disturbance covariance
global pvtol_Rw;			% Noise covariance
global pvtol_C;				% outputs to use
global pvtol_F;				% disturbance input

% Parametro del sistema
J = 0.0475;			% inertia around pitch axis
m = 1.5;			% mass of fan
r = 0.25;			% distance to flaps
g = 10;				% gravitational constant
d = 0.2;	 		% damping factor (estimated)

% Inicializar las derivadas al tamano correcto
deriv = zeros(size(state));

% Extraer los estados estimados actuales
x = state(1:6);
xhat = state(7:12);

% Obtener la salida actual, con ruido
y = pvtol_C*x + pvtol_C * ...
  [0.1*sin(2.1*t); 0.1*sin(3.2*t); 0; 0; 0; 0];

% Calcular las fuerzas de disturbio
fd = [  0.01*sin(0.1*t); 0.01*cos(0.027*t); 0];

% Calcular la ley de control
F = -pvtol_K * xhat + [0; m*g];

% Matriz A con estados estimados actuales
A = [  0     0     0     1     0     0;
       0     0     0     0     1     0;
       0     0     0     0     0     1;
       0, 0, (-F(1)*sin(xhat(3)) - F(2)*cos(xhat(3)))/m, -d, 0, 0;
       0, 0, (F(1)*cos(xhat(3)) - F(2)*sin(xhat(3)))/m, 0, -d, 0;
       0     0     0     0     0     0 ];

% Dinamica del estimador dynamics (prediccion)
deriv(7) = xhat(4); deriv(8) = xhat(5); deriv(9) = xhat(6);
deriv(10) = (F(1) * cos(xhat(3)) - F(2) * sin(xhat(3)) - d*xhat(4)) / m;
deriv(11) = (F(1) * sin(xhat(3)) + F(2) * cos(xhat(3)) - m*g - d*xhat(5)) / m;
deriv(12) = (F(1) * r) / J;

% Calcular la covarianza
P = reshape(state(13:48), 6, 6);
dP = A * P + P * A' - P * pvtol_C' * inv(pvtol_Rv) * pvtol_C * P + ...
  pvtol_F * pvtol_Rw * pvtol_F';
L = P * pvtol_C' * inv(pvtol_Rv);

% Calcular el termino de correccion
xcor = L * (y - pvtol_C*xhat);
for i = 1:6, deriv(6+i) = deriv(6+i) + xcor(i); end;

% dinamica del ventilador de flujo guiado
deriv(1) = x(4); deriv(2) = x(5); deriv(3) = x(6);
deriv(4) = (F(1)*cos(x(3)) - F(2)*sin(x(3)) - d*x(4) + fd(1)) / m;
deriv(5) = (F(1)*sin(x(3)) + F(2)*cos(x(3)) - m*g - d*x(5) + fd(2)) / m;
deriv(6) = (F(1) * r + fd(3)) / J;

% Copiar en las actualizaciones de covarianzas
for i = 1:6,
  for j = 1:6,
    deriv(6+6*i+j) = dP(i, j);
  end;
end;

% All done
return;
